###########################################################
#
# General make-include
# Johan Bezem, JB Enterprises, (c) 2008.
#
# Prerequisites:
#  - GNUmake   3.79.1 or higher
#  - Bash      2.05a or higher
#    or
#  - CMD.EXE   Windows XP or higher
#
# Version 003.000
# Saved under "003Sources.zip"

# Provide a different version of 'mkdir' depending on the platform
ifeq (1,$(YT_YN_USE_CMDEXE))
  makedirectory=if not exist $(1) md $(subst /,\\,$(1))
else
  makedirectory=if [ ! -d $(1) ]; then mkdir $(1); fi
endif

# Since 'make' is called recursively, every line in this Makefile
# will get executed again every time, in every directory anew.
# Therefore, we'll create a segment where several variables are 
# determined and exported, since exporting them already makes them 
# available to all child processes.
ifeq ($(MAKELEVEL),0)

  # First, determine the name of the objects directory
  # This name should be in such a way unique, that a pathname
  # containing that string can only refer to an objects resp.
  # intermediates' directory. This is of vital importance for
  # the structure of the makefile.
  export YT_OBJDIRNAME:=objects

endif

ifeq (,$(findstring $(YT_OBJDIRNAME),$(CURDIR)))
  YT_OBJDIR := $(YT_PBASE)/$(YT_OBJDIRNAME)

  # Disable all built-in rules; we don't need them on this run
  .SUFFIXES:

  # The object directory is the only target here; since it's phony,
  # when the commands have been executed, it is considered built.
  # This calls make recursively, but now from the objects directory.
  .PHONY: $(YT_OBJDIR)
  $(YT_OBJDIR):
	+@$(call makedirectory,$@)
	+@echo Make : $@
	+@$(MAKE) -C $@ -f $(CURDIR)/Makefile YT_SRCVPATH=$(CURDIR) YT_OBJBASE=$(YT_OBJDIR) --no-print-directory $(MAKECMDGOALS)

  # Since we'll provide just one rule, but a 'match-anything' rule,
  # when make tries to remake the makefiles involved, the rule will
  # also apply, and make would recursively call itself from the
  # same directory... Endless recursion results in chaos.
  Makefile : ; @:
  %.mk :: ;

  # No matter what goals are given, this dependency will make sure,
  # make is recursively called.
  % :: $(YT_OBJDIR) ; @:

else

  # Compile the include path to refer to all include
  # directories recursively; use $(wildcard to filter-out
  # non-existing directories
  export YT_INCLUDEPATH := $(strip $(wildcard $(YT_SRCVPATH)/include) $(YT_INCLUDEPATH))

  # Now we can add the vpath directive for include files
  # (Add other extensions as necessary). This vpath is only necessary
  # to find/recognize the various *.h files specified as dependencies
  # by the call to maxdepend, generating a set of *.d files that are
  # included at the bottom of this file.
  # See the definition of YT_CPPFLAGS for a description of all
  # used directories.
  vpath %.h $(YT_SRCVPATH) $(YT_INCLUDEPATH)

  # Now set the various option flags; use YT_ as prefix
  # to avoid name clashes with the default
  # variable names
  YT_CPPFLAGS := $(patsubst %,-I%,$(YT_INCLUDEPATH))
  YT_CFLAGS   := $(CFLAGS)

  # Determine all sources, objects, etc.
  # Be sure not to retain any pathnames
  YT_SOURCES:=$(notdir $(wildcard $(YT_SRCVPATH)/*.c))

  # Create the objects' filenames from the source filenames
  YT_OBJECTS := $(patsubst %.c,%.obj,$(YT_SOURCES))

  # Default target
  .PHONY: all
  all: RECURSE $(YT_OBJECTS)

  # Clean target
  .PHONY: clean
  clean: RECURSE
	@rm -f *.obj

  # List_dirs will contain a space delimited list of all directories
  # containing a Makefile.
  list_dirs := $(dir $(wildcard $(YT_SRCVPATH)/*/Makefile))
  # patsubst retains only the directory names, not the full paths
  list_dirs := $(patsubst $(YT_SRCVPATH)/%,%,$(list_dirs))

  # We define RECURSE as PHONY here, so we don't need to specify it in
  # case we don't need recursion anymore (the else-clause for the following
  # ifneq, when no more subdirectories containing a Makefile can be found).
  # The directories are to be phony, in order to execute the
  # recursion commands for all directories in all cases.
  .PHONY: RECURSE $(list_dirs)

  # If list_dirs is not empty, we need to recurse through one or more 
  # subdirectories
  ifneq ($(strip $(list_dirs)),)

    # The RECURSE target can be used as a dependency for all targets
    # that need to be made recursively. Put it as the first dependency
    # for a depth-first usage.
    RECURSE: $(list_dirs)

    # To 'create' the subdirectories, we execute the commands listed,
    # one of which is reexecuting 'make' in the target subdirectory '$@'
    $(list_dirs):
	+@$(call makedirectory,$(YT_OBJBASE)/$@)
	+@echo Make[$(MAKELEVEL)]: $(patsubst %/,%,$@)
	+@$(MAKE) -C $(YT_OBJBASE)/$(patsubst %/,%,$@) YT_SRCVPATH=$(YT_SRCVPATH)/$(patsubst %/,%,$@) YT_OBJBASE=$(YT_OBJBASE)/$(patsubst %/,%,$@) -f $(YT_SRCVPATH)/$(patsubst %/,%,$@)/Makefile --no-print-directory $(MAKECMDGOALS)
    # Watch the empty line before the endif, otherwise it would be an (illegal)
    # part of the commands to $(list_dirs): !!

  endif

  # Rules are mostly self-defined, since different compilers
  # have different customs. So here we clear the list of implicit
  # pattern rules and known suffixes.
  .SUFFIXES:
  # A .SUFFIXES: rule with extensions apparently only means anything 
  # to old style suffix rules, but we define it anyway
  # For now we only recognize C-files.
  .SUFFIXES: .c

  vpath %.c $(YT_SRCVPATH)

  %.obj: %.c
	@echo .c   to $(YT_OBJEXT): $(notdir $<)
	$(YT_S)gcc -c $(YT_CPPFLAGS) $(YT_CFLAGS) $< -o $@

endif
